from django.conf import settings

HOSTNAME = getattr(settings, 'HOSTNAME', '')
TERRA_TILES_HOSTNAMES = getattr(settings, 'TERRA_TILES_HOSTNAMES', [HOSTNAME, ])

SWAGGER_ENABLED = getattr(settings, 'SWAGGER_ENABLED', False)

MAX_TILE_ZOOM = getattr(settings, 'MAX_TILE_ZOOM', 15)
MIN_TILE_ZOOM = getattr(settings, 'MIN_TILE_ZOOM', 10)

INTERNAL_GEOMETRY_SRID = getattr(settings, 'INTERNAL_GEOMETRY_SRID', 4326)



import hashlib
from random import uniform

import mercantile
from django.core.cache import cache
from django.db import connection
from django.db.models import F
from math import ceil, floor, log, pi

from . import EARTH_RADIUS, EPSG_3857
from .funcs import (ST_Area, ST_Length, ST_MakeEnvelope,
                    ST_SimplifyPreserveTopology, ST_Transform)
from .sigtools import SIGTools
from .. import settings as app_settings


def get_cache_version(layer):
    try:
        return int(layer.features.order_by('-updated_at').first().updated_at.timestamp())
    except AttributeError:
        # This happens when a layer is empty
        return 1


def cached_tile(func):
    def wrapper(self, x, y, z,
                pixel_buffer, features_filter, properties_filter,
                features_limit, *args, **kwargs):
        cache_key = self.get_tile_cache_key(
            x, y, z,
            pixel_buffer, features_filter, properties_filter,
            features_limit)

        version = get_cache_version(self.layer)

        expiration_factor = (log(5, z) ** 0.9)

        # Cache expiry calculation is based on a logarithmic function where we add a random factor of +-10%
        expiration = int(expiration_factor * (3600 * 24 * 7) * uniform(0.9, 1.1))

        def build_tile():
            (a, b) = func(
                self, x, y, z,
                pixel_buffer, features_filter, properties_filter,
                features_limit, *args, **kwargs)
            return (a, b.tobytes())

        return cache.get_or_set(cache_key, build_tile, expiration, version=version)

    return wrapper


class VectorTile(object):
    def __init__(self, layer, cache_key=None):
        self.layer, self.cache_key = layer, cache_key

    # Number of tile units per pixel
    EXTENT_RATIO = 8
    TILE_WIDTH_PIXEL = 512

    def _simplify(self, layer_query, pixel_width_x, pixel_width_y):
        if self.layer.is_polygon:
            # Grid step is pixel_width_x / EXTENT_RATIO and pixel_width_y / EXTENT_RATIO
            # Simplify to average half pixel width
            layer_query = layer_query.annotate(
                outgeom3857=ST_SimplifyPreserveTopology('outgeom3857', (pixel_width_x + pixel_width_y) / 2 / 2)
            )
        return layer_query

    def _filter_on_property(self, layer_query, features_filter):
        if features_filter is not None:
            layer_query = layer_query.filter(
                properties__contains=features_filter
            )
        return layer_query

    def _filter_on_geom_size(self, layer_query, layer_geometry, pixel_width_x, pixel_width_y):
        if self.layer.is_linestring:
            # Larger then a half of pixel
            layer_query = layer_query.annotate(
                length3857=ST_Length('outgeom3857')
            ).filter(
                length3857__gt=(pixel_width_x + pixel_width_y) / 2 / 2
            )
        elif self.layer.is_polygon:
            # Larger than a quarter of pixel
            layer_query = layer_query.annotate(
                area3857=ST_Area('outgeom3857')
            ).filter(
                area3857__gt=pixel_width_x * pixel_width_y / 4
            )
        return layer_query

    def _limit(self, layer_query, features_limit):
        if features_limit is not None:
            # Order by feature size before limit
            if self.layer.is_linestring:
                layer_query = layer_query.order_by('length3857')
            elif self.layer.is_polygon:
                layer_query = layer_query.order_by('area3857')

            layer_query = layer_query[:features_limit]
        return layer_query

    def get_tile_bbox(self, x, y, z):
        bounds = mercantile.bounds(x, y, z)
        return (*mercantile.xy(bounds.west, bounds.south), *mercantile.xy(bounds.east, bounds.north))

    def pixel_widths(self, xmin, ymin, xmax, ymax):
        return (
            (xmax - xmin) / self.TILE_WIDTH_PIXEL,
            (ymax - ymin) / self.TILE_WIDTH_PIXEL
        )

    @cached_tile
    def get_tile(self, x, y, z,
                 pixel_buffer, features_filter, properties_filter,
                 features_limit, features):

        xmin, ymin, xmax, ymax = self.get_tile_bbox(x, y, z)
        pixel_width_x, pixel_width_y = self.pixel_widths(xmin, ymin, xmax, ymax)

        layer_query = features.annotate(
            bbox=ST_MakeEnvelope(
                xmin,
                ymin,
                xmax,
                ymax,
                EPSG_3857),
            # Intersects on internal data projection using pixel buffer
            bbox_select=ST_Transform(ST_MakeEnvelope(
                xmin - pixel_width_x * pixel_buffer,
                ymin - pixel_width_y * pixel_buffer,
                xmax + pixel_width_x * pixel_buffer,
                ymax + pixel_width_y * pixel_buffer,
                EPSG_3857), app_settings.INTERNAL_GEOMETRY_SRID),
            outgeom3857=ST_Transform('geom', EPSG_3857),
        ).filter(
            bbox_select__intersects=F('geom')
        )

        # Filter features
        layer_query = self._filter_on_property(layer_query, features_filter)
        layer_query = self._filter_on_geom_size(layer_query, self.layer.layer_geometry, pixel_width_x, pixel_width_y)

        # Lighten geometry
        layer_query = self._simplify(layer_query, pixel_width_x, pixel_width_y)

        # Seatbelt
        layer_query = self._limit(layer_query, features_limit)

        layer_raw_query, args = layer_query.query.sql_with_params()

        if properties_filter:
            filter = ', '.join([f"'{f}'" for f in properties_filter])
            properties = f'''
                (
                    SELECT jsonb_object_agg(key, value) FROM
                    jsonb_each(properties)
                    WHERE key IN ({filter})
                )
                '''
        elif properties_filter == []:
            properties = "'{}'::jsonb"
        else:
            properties = "properties"

        properties += " || json_build_object('_id', identifier)::jsonb"

        with connection.cursor() as cursor:
            sql_query = f'''
                WITH
                fullgeom AS ({layer_raw_query}),
                tilegeom AS (
                    SELECT
                        ({properties}) AS properties,
                        ST_AsMvtGeom(
                            outgeom3857,
                            bbox,
                            {self.TILE_WIDTH_PIXEL * self.EXTENT_RATIO},
                            {pixel_buffer * self.EXTENT_RATIO},
                            true) AS geometry
                    FROM
                        fullgeom)
                SELECT
                    count(*) AS count,
                    ST_AsMVT(
                        tilegeom,
                        CAST(%s AS text),
                        {self.TILE_WIDTH_PIXEL * self.EXTENT_RATIO},
                        'geometry'
                    ) AS mvt
                FROM
                    tilegeom
            '''

            cursor.execute(sql_query, args + (self.layer.name,))
            row = cursor.fetchone()

            return row[0], row[1]

    def get_tile_cache_key(self, x, y, z,
                           pixel_buffer, features_filter, properties_filter,
                           features_limit):
        if self.cache_key:
            cache_key = self.cache_key
        else:
            cache_key = self.layer.pk
        features_filter = ''
        if features_filter is not None:
            features_filter_hash = \
                hashlib.sha224(
                    str(features_filter).encode('utf-8')
                ).hexdigest()
        properties_filter_hash = ''
        if properties_filter is not None:
            properties_filter_hash = \
                hashlib.sha224(
                    ','.join(properties_filter).encode('utf-8')
                ).hexdigest()
        return (
            f'tile_cache_{cache_key}_{x}_{y}_{z}'
            f'_{pixel_buffer}_{features_filter_hash}_{properties_filter_hash}'
            f'_{features_limit}'
        )

    def clean_tiles(self, tiles, pixel_buffer, features_filter,
                    properties_filter, features_limit):
        """
        TODO Cleaning tiles must be rewrited but for the moment invalidating tiles is not necessary
        """
        pass


def guess_maxzoom(layer):
    features = layer.features.all()
    layer_query = features.annotate(
        geom3857=ST_Transform('geom', EPSG_3857)
    )

    layer_raw_query, args = layer_query.query.sql_with_params()

    try:
        with connection.cursor() as cursor:
            sql_query = f'''
                WITH
                q1 AS ({layer_raw_query}),
                q2 AS (SELECT ST_X((ST_DumpPoints(geom3857)).geom) AS x FROM q1),
                q3 AS (SELECT x - lag(x) OVER (ORDER BY x) AS dst FROM q2),
                q4 AS (SELECT * FROM q3 WHERE dst > 0)
                SELECT
                    exp(sum(ln(dst))/count(dst)) AS avg
                FROM
                    q4
                '''
            cursor.execute(sql_query, args)
            row = cursor.fetchone()

        # geometric mean of the |x_{i+1}-x_i| for all points (x,y) in layer.pk
        avg = row[0]

        # total number of pixels to represent length `avg` (equator)
        nb_pixels_total = 2 * pi * EARTH_RADIUS / avg

        tile_resolution = VectorTile.TILE_WIDTH_PIXEL * VectorTile.EXTENT_RATIO

        # zoom (ceil) to fit those pixels at `tile_resolution`
        max_zoom = ceil(log(nb_pixels_total / tile_resolution, 2))

        return min(max_zoom, 22)
    except TypeError:
        return 14  # Arbitrary zoom value


def guess_minzoom(layer):
    """
    Procedure that uses DB output to guess a min zoom level.

    Criteria: zoom satisfying the following condition:
    tile_lenght / tile_fraction <= BBox(features).smallerSide
    If extent = 0, returns 0

    Explanation about the tile_fraction = 8 just above:
    ---------------------------------------------------
    It's purpose is to give an idea of when a dataset becomes to small
    to be shown in the map, so when doing.
    """

    extent = SIGTools.get_extent_of_layer(layer)

    if extent == 0:
        return 0

    tile_fraction = 8
    min_zoom = floor(log((2 * pi * EARTH_RADIUS) / (extent * tile_fraction), 2))
    return min(min_zoom, 22)



import json
from urllib.parse import unquote, urljoin

from django.core.cache import cache
from django.http import HttpResponse
from django.urls import reverse
from django.utils.encoding import escape_uri_path
from django.utils.html import escape

from ..models import Feature, Layer
from .. import settings as app_settings
from .helpers import VectorTile


class AbstractTileJsonMixin:
    response_class = HttpResponse
    content_type = 'application/json'

    @staticmethod
    def settings_link(layer, *args):
        layer_settings = layer.layer_settings_with_default(*args)
        if 'link' in layer_settings:
            return '<a href="{0}"/>{1}</a>'.format(
                escape_uri_path(layer_settings['link']),
                escape(layer_settings['name'])
            )
        return layer_settings

    @property
    def layers(self):
        raise NotImplementedError()

    def get_tile_path(self):
        raise NotImplementedError()

    def get_minzoom(self):
        return max(
            app_settings.MIN_TILE_ZOOM,
            min([
                l.layer_settings_with_default('tiles', 'minzoom')
                for l in self.layers
            ]))

    def get_maxzoom(self):
        return min(
            app_settings.MAX_TILE_ZOOM,
            max([
                l.layer_settings_with_default('tiles', 'maxzoom')
                for l in self.layers
            ]))

    def _join_group_settings_link(self, layers, *args):
        return ','.join(set([
            a if 'link' not in a else
            '<a href="{0}"/>{1}</a>'.format(escape_uri_path(a['link']), escape(a['name']))
            for a in [layer.layer_settings_with_default(*args) for layer in self.layers]
            if a
        ])) or None

    def _join_group_settings_string(self, layers, *args):
        return ','.join(set([
            a
            for a in [layer.layer_settings_with_default(*args) for layer in self.layers]
            if a
        ])) or None

    def get_attribution(self):
        return self._join_group_settings_link(self.layers, 'metadata', 'attribution')

    def get_description(self):
        return self._join_group_settings_string(self.layers, 'metadata', 'description')

    @staticmethod
    def layer_fields(layer):
        properties_filter = layer.layer_settings_with_default(
            'tiles', 'properties_filter')
        if properties_filter is not None:
            fields = properties_filter
        else:
            fields = layer.layer_properties.keys()

        return {f: '' for f in fields}

    def get_vector_layers(self):
        return [
            {
                'id': l.name,
                'fields': self.layer_fields(l),
                'minzoom': l.layer_settings_with_default('tiles', 'minzoom'),
                'maxzoom': l.layer_settings_with_default('tiles', 'maxzoom'),
            } for l in self.layers]

    def get_tilejson(self):
        minzoom = self.get_minzoom()
        maxzoom = self.get_maxzoom()

        tile_path = self.get_tile_path()

        # https://github.com/mapbox/tilejson-spec/tree/3.0/3.0.0
        return {
            'tilejson': '3.0.0',
            'name': self.object.name,
            'tiles': [
                unquote(urljoin(hostname, tile_path))
                for hostname in app_settings.TERRA_TILES_HOSTNAMES
            ],
            'minzoom': minzoom,
            'maxzoom': maxzoom,
            # bounds
            # center
            'attribution': self.get_attribution(),
            'description': self.get_description(),
            'vector_layers': self.get_vector_layers(),
        }

    def get_last_update(self):
        features = Feature.objects.filter(layer__in=self.layers).order_by('-updated_at')

        if features.exists():
            ref_object = features.first()
        else:
            ref_object = Layer.objects.filter(pk__in=self.layers).order_by('-updated_at').first()

        return ref_object.updated_at

    def render_to_response(self, context, **response_kwargs):
        self.object = self.get_object()
        last_update = self.get_last_update()

        cache_key = f'tilejson-{self.object.name}'
        version = int(last_update.timestamp())
        tilejson_data = cache.get(cache_key, version=version)

        if tilejson_data is None:
            tilejson_data = json.dumps(self.get_tilejson())
            cache.set(cache_key, tilejson_data, version=version)

        response_kwargs.setdefault('content_type', self.content_type)
        return self.response_class(
            content=tilejson_data,
            **response_kwargs,
        )


class TileJsonMixin(AbstractTileJsonMixin):
    def get_tile_path(self):
        return reverse("geostore:layer-tiles-pattern", args=[self.object.pk])

    @property
    def layers(self):
        # keep a qs result here
        return type(self.object).objects.filter(pk=self.object.pk)


class MultipleTileJsonMixin(AbstractTileJsonMixin):
    def get_tile_path(self):
        return reverse("geostore:group-tiles-pattern", args=[self.object.slug])

    @property
    def layers(self):
        # keep a qs result here
        return self.object.layers.all()


class TileResponseMixin:
    response_class = HttpResponse
    content_type = 'application/vnd.mapbox-vector-tile'

    def get_tile_for_layer(self, layer):
        tile = VectorTile(layer)
        features = layer.features.all()
        return tile.get_tile(
            self.kwargs['x'], self.kwargs['y'], self.kwargs['z'],
            layer.layer_settings_with_default('tiles', 'pixel_buffer'),
            layer.layer_settings_with_default('tiles', 'features_filter'),
            layer.layer_settings_with_default('tiles', 'properties_filter'),
            layer.layer_settings_with_default('tiles', 'features_limit'),
            features,
        )

    def get_tile(self, layer):
        minzoom = layer.layer_settings_with_default('tiles', 'minzoom')
        maxzoom = layer.layer_settings_with_default('tiles', 'maxzoom')
        if self.kwargs['z'] >= minzoom and self.kwargs['z'] <= maxzoom:
            _, tile = self.get_tile_for_layer(layer)
            return tile
        return b''

    def render_to_response(self, context, **response_kwargs):
        layer = context['object']
        tile = self.get_tile(layer)

        response_kwargs.setdefault('content_type', self.content_type)
        return self.response_class(
            content=tile,
            **response_kwargs,
        )


class MultipleTileResponseMixin(TileResponseMixin):

    def get_tile(self, layers):
        return b''.join([
            super(MultipleTileResponseMixin, self).get_tile(layer)
            for layer in layers
        ])

    def render_to_response(self, context, **response_kwargs):
        layers = context['object'].layers.all()
        tile = self.get_tile(layers)

        response_kwargs.setdefault('content_type', self.content_type)
        return self.response_class(
            content=tile,
            **response_kwargs,
        )



# Generated by Django 2.0.13 on 2019-02-28 16:55

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0017_auto_20181114_0806'),
    ]

    operations = [
        migrations.AlterField(
            model_name='feature',
            name='properties',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict),
        ),
    ]



from django.db import migrations, models


def create_groups(apps, schema_editor):
    LayerGroup = apps.get_model('geostore', 'LayerGroup')
    LayerGroup.objects.create(name='default', slug='default')
    Layer = apps.get_model('geostore', 'Layer')
    for layer in Layer.objects.all():
        group_name = layer.group or 'default'
        group, created = LayerGroup.objects.get_or_create(name=group_name, defaults={'slug': group_name})
        group.layers.add(layer)


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0026_auto_20190613_1529'),
    ]

    operations = [
        migrations.CreateModel(
            name='LayerGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True,
                                        serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=256, unique=True)),
                ('slug', models.SlugField(unique=True)),
                ('layers', models.ManyToManyField(to='geostore.Layer', related_name='layer_groups')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.RunPython(create_groups),
    ]



from django.contrib.gis.db.models import Extent
from django.core.management.base import BaseCommand
from django.db import transaction
from mercantile import tiles

from geostore.models import Layer
from geostore.tiles.helpers import VectorTile


class Command(BaseCommand):
    help = 'Generate tiles cache of all layers'

    @transaction.atomic()
    def handle(self, *args, **options):
        for layer in Layer.objects.all():
            if options['verbosity'] >= 1:
                self.stdout.write(f'Generating {layer.name} tiles cache')
            bbox = layer.features.aggregate(bbox=Extent('geom'))['bbox']
            if bbox:
                vtile = VectorTile(layer)
                zoom_range = range(
                    layer.layer_settings_with_default('tiles', 'minzoom'),
                    layer.layer_settings_with_default('tiles', 'maxzoom') + 1
                )

                for tile in tiles(*bbox, zoom_range):
                    vtile.get_tile(
                        tile.x, tile.y, tile.z,
                        layer.layer_settings_with_default('tiles', 'pixel_buffer'),
                        layer.layer_settings_with_default('tiles', 'features_filter'),
                        layer.layer_settings_with_default('tiles', 'properties_filter'),
                        layer.layer_settings_with_default('tiles', 'features_limit'),
                        layer.features.all())



from django.views.generic.detail import BaseDetailView

from ..models import Layer, LayerGroup
from .mixins import (MultipleTileJsonMixin, MultipleTileResponseMixin,
                     TileJsonMixin, TileResponseMixin)


class TileJsonView(TileJsonMixin, BaseDetailView):
    model = Layer


class MultipleTileJsonView(MultipleTileJsonMixin, BaseDetailView):
    model = LayerGroup


class LayerTileDetailView(TileResponseMixin, BaseDetailView):
    model = Layer


class LayerGroupTileDetailView(MultipleTileResponseMixin, BaseDetailView):
    model = LayerGroup
